#!/usr/bin/env python3

# Copyright (C) 2023:
#   Helmholtz-Zentrum Potsdam Deutsches GeoForschungsZentrum GFZ
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
# General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.

import copy
import logging
import datetime
import time

from multiprocessing import Queue, Pool
from rulelib import RuleHandler
from databaselib import SpatiaLiteDatabase

logger = logging.getLogger(__name__)


class BuildingInitializer:
    """
    The `BuildingInitializer` class copies and enriches building data from an OpenStreetMap
    replication database.

    Args:
        rules (dict):
            A dictionary with rule names as keys and definitions as values.
        osm_replication_db (PostGISDatabase):
            The `OSM replication` database with the source data.
        exposure_db (AbstractDatabase):
            The output database with processed building information.

    Attributes:
        rule_handler (RuleHandler):
            Evaluates given rules based on a set of arguments.
        osm_replication_db (PostGISDatabase):
            The `OSM replication` database with the source data.
        exposure_db (AbstractDatabase):
            The output database with processed building information.
    """

    building_source_db_name: str = "source"

    def __init__(
        self,
        rules,
        osm_replication_db,
        exposure_db,
        source_db,
        building_table,
        building_id_column,
        building_geometry_column,
    ):
        self.rule_handler = RuleHandler(rules)
        self.osm_replication_db = osm_replication_db
        self.exposure_db = exposure_db
        self.source_db = source_db
        self.building_table = building_table
        self.building_id_column = building_id_column
        self.building_geometry_column = building_geometry_column

        self.outputs = set()
        for rule in self.rule_handler:
            self.outputs.update(rule.outputs)

        if osm_replication_db is None:
            raise ValueError("No source database object is found.")
        if exposure_db is None:
            raise ValueError("No building database object is found.")

    def retrieve_building_ids(
        self,
        sql_statement=None,
        building_id_column=None,
        building_table=None,
        building_geometry_column=None,
    ):
        """
        Retrieve a selection of building IDs from the building table, using a custom SQL
        statement. If no statement is given, all IDs are retrieved. Without an SQL statement
        provided, the `building_id_column`, `building_table` and `building_geometry_column`
        are used to create an SQL statement. If these are not given, the class variables are
        used instead.

        Args:
            sql_statement (str, optional, default: None):
                The SQL statement used to retrieve the building IDs.
            building_id_column (str, optional, default: None):
                ID column name of the building table.
            building_table (str, optional, default: None):
                Name of the building table.
            building_geometry_column (str, optional, default: None):
                Geometry column name of the building table.
        """

        # Create default SQL query if not given as function parameter.
        if sql_statement is None:
            if building_id_column is None:
                building_id_column = self.building_id_column
            if building_table is None:
                building_table = self.building_table
            if building_geometry_column is None:
                building_geometry_column = self.building_geometry_column

            sql_statement = f"""
                SELECT {building_id_column}
                FROM {building_table}
                ORDER BY ST_YMAX({building_geometry_column}) DESC
                """

        # Execute SQL statement on the building source database.
        if self.building_source_db_name == "osm_replication":
            building_source_db = self.osm_replication_db
        else:
            building_source_db = self.source_db

        building_source_db.connect()
        logger.info(f"Retrieving all building IDs using the SQL statement:\n{sql_statement}")
        building_source_db.cursor.execute(sql_statement)
        building_ids = [building_id for building_id, in building_source_db.cursor]
        logger.info(f"Retrieved {len(building_ids)} building IDs.")
        building_source_db.close()
        return building_ids

    def run(self, building_ids, batch_size=500):
        """
        Process a list of building IDs, based on the rules defined during initialization. For
        each batch, the building information is retrieved, the information is evaluated
        according to the rules and the output is inserted or updated to the building database.

        Args:
            building_ids (list):
                A list of building IDs that should be processed.
            batch_size (int, default: 500):
                The number of features that should be processed per batch.
        """

        # Connect to the databases.
        self.osm_replication_db.connect()
        self.exposure_db.connect()
        self.source_db.connect()

        if self.building_source_db_name == "osm_replication":
            building_source_db = self.osm_replication_db
        else:
            building_source_db = self.source_db

        for _building_ids in self.batch_iter(building_ids, batch_size):
            # Get the building information for a batch of building IDs.
            building_information = self.get_building_information(
                _building_ids, building_source_db.cursor
            )

            # Get the building information about its surroundings for a batch of OSM IDs.
            building_information = self.get_additional_information(
                _building_ids, building_information, self.osm_replication_db.cursor
            )

            # Get the most overlapping GHSL characteristics types.
            building_information = self.get_ghsl_characteristics_type(
                _building_ids, building_information, self.source_db.cursor
            )

            # Evaluate the building information and retrieve a dictionary with the output
            # information.
            output = self.evaluate_building_information(building_information)

            # Insert or update the building database with the output.
            self.upsert(output, self.exposure_db.cursor)
            self.exposure_db.connection.commit()

        # Close the database connections.
        self.exposure_db.close()
        self.osm_replication_db.close()

    def run_multiprocess(self, building_ids, batch_size=500, num_processors=5):
        """
        Fills a queue with building IDs in batches. Creates a pool of workers that process the
        batches.

        Args:
            building_ids (list):
                A list of building IDs that should be processed.
            batch_size (int, default: 500):
                The number of features that should be processed per batch.
            num_processors (int):
                Number of workers that are created to run the processes.
        """

        # Do not allow multiprocessing with a SpatiaLite database.
        if isinstance(self.exposure_db, SpatiaLiteDatabase):
            raise TypeError(
                "The multiprocessing functionality is not available with SpatiaLite"
            )

        start_process = datetime.datetime.now()

        # Setup queue and processing pool
        queue = Queue()
        pool = Pool(num_processors, self.worker, (queue,))

        # Batch iterate through the building IDs and put the batches into the queue
        n = 0
        for n, _building_ids in enumerate(self.batch_iter(building_ids, batch_size)):
            queue.put(_building_ids)

        # Wait until queue is empty and meanwhile log information
        logger.info(f"{n + 1} batches are in the queue")
        while queue.qsize() > 0:
            time_spent = (datetime.datetime.now() - start_process).total_seconds()
            logger.info(
                f"Length queue: {queue.qsize()}\tTotal time spent: {time_spent} seconds"
            )
            time.sleep(10)

        # Stop workers and close pool
        for i in range(num_processors):
            queue.put(None)

        # Close the pool
        pool.close()
        # Wait for all processes to finish, before the main thread continues
        pool.join()

        time_spent = (datetime.datetime.now() - start_process).total_seconds()
        logger.info(
            f"Processed {n + 1} batches in {time_spent} seconds on {num_processors} cores"
        )

    def worker(self, queue):
        """
        Process batches of building IDs on one core, based on the rules defined during
        initialization. For each batch, the building information is retrieved, the information
        is evaluated according to the rules and the output is inserted or updated to the
        building database.

        Args:
            queue:
                The multiprocessing queue with all building ID batches.
        """

        # Copy source- and building database objects and connect to the database.
        osm_replication_db, source_db, exposure_db = self.copy_postgis_database_objects()

        osm_replication_db.connect()
        exposure_db.connect()
        source_db.connect()

        # For OSM buildings, the building source is the `osm_replication_db`, while for all
        # other building sources, it is the `source_db`.
        if self.building_source_db_name == "osm_replication":
            building_source_db = osm_replication_db
        else:
            building_source_db = source_db

        while True:
            building_ids = queue.get()
            if building_ids is None:
                break

            # Get the building information for a batch of building IDs.
            building_information = self.get_building_information(
                building_ids, building_source_db.cursor
            )

            # Get the building information about its surroundings for a batch of OSM IDs.
            building_information = self.get_additional_information(
                building_ids, building_information, osm_replication_db.cursor
            )

            # Get the most overlapping GHSL characteristics types.
            building_information = self.get_ghsl_characteristics_type(
                building_ids, building_information, source_db.cursor
            )

            # Evaluate the building information and retrieve a dictionary with the output
            # information.
            output = self.evaluate_building_information(building_information)

            # Insert or update the building database with the output.
            self.upsert(output, exposure_db.cursor)
            exposure_db.connection.commit()

        # Close connections.
        osm_replication_db.close()
        exposure_db.close()

    def evaluate_building_information(self, building_information: dict):
        """
        Given a dictionary with building IDs as keys and building information as values,
        evaluates the set of rules that are included in the `rule_handler`.

        Args:
            building_information (dict):
                A dictionary with building IDs as keys and building information as values.

        Returns:
            The output of the given set of rules in a dictionary with the building IDs as keys.
        """

        result = {}
        for building_id, args in building_information.items():
            result[building_id] = self.rule_handler.eval(args)
        return result

    def upsert(self, result, cursor):
        """
        Insert or update the resulting building information in the building database.

        Args:
            result (dict):
                Dictionary with building IDs as keys and the resulting building information as
                values. The building information is a dictionary with column names as keys and
                the information as values.
            cursor (PostGISDatabase.cursor):
                Connected cursor to the OSM Replication database.
        """

        # The rule outputs define the columns that should be upserted.
        columns = list(self.outputs)

        # Format values so that they can be uploaded as a batch on a PostGIS database.
        values = []
        for building_id, params in result.items():
            _values = ",".join([self.format_value(params[key], key) for key in columns])
            values.append(f"({building_id},{_values})")

        # Set conflict values to the new values, if the building ID is already there
        on_conflict = ",".join([f"{k} = EXCLUDED.{k}" for k in columns])
        # Format column headers
        columns = ",".join(columns)
        # Format values
        values = ",".join(values)
        # Add all relevant columns to the database. If the building ID already exists, update
        # the attributes.
        sql_statement = f"""
            INSERT INTO {self.building_table} ({self.building_id_column}, {columns})
            VALUES {values}
            ON CONFLICT ({self.building_id_column}) DO UPDATE
            SET {on_conflict}
            """

        cursor.execute(sql_statement)

    @staticmethod
    def format_value(value, key):
        """
        Format values correctly to be used in an SQL statement.

        Args:
            value (object):
                The value itself.
            key (str):
                Name of the column in the database.

        Returns:
            Correctly formatted value.
        """

        # None values should be `NULL` in PostgreSQL
        if value is None:
            return "NULL"
        # The geometry is already correctly formatted
        elif key in ["geometry", "geom"]:
            return value
        # If a value is a string, it should be in quotation marks
        elif isinstance(value, str):
            return f"'{value}'"
        # Else return a string representation of the value
        else:
            return str(value)

    def get_building_information(self, building_ids, cursor):
        """
        Get the information of a batch of buildings, including the attributes of the building,
        the geometry of the building and all information related to its geometry.

        Args:
            building_ids (list):
                List of building IDs.
            cursor (PostGISDatabase.cursor):
                Connected cursor to the OSM Replication database.

        Returns:
            A dictionary with the building IDs as keys and the building information as values.
        """

        if not isinstance(building_ids, str):
            building_ids = ",".join([str(idx) for idx in building_ids])

        building_information = {}

        sql_statement = self.get_building_information_query(building_ids)
        cursor.execute(sql_statement)
        for result in cursor:
            _building_information = {k.name: v for k, v in zip(cursor.description, result)}
            if "tags" in _building_information:
                _building_information["tags"] = dict(_building_information["tags"])
            else:
                _building_information["tags"] = {}

            # Add the `height` attribute to the building tags, so all building sources get
            # parsed with the same rules as the OpenStreetMap buildings.
            if "height" in _building_information:
                _building_information["tags"]["height"] = _building_information["height"]

            building_information[_building_information[self.building_id_column]] = {
                **_building_information,
                "relations": [],
                "osm_spots": [],
                "osm_lands": [],
            }
        return building_information

    def get_additional_information(self, building_ids, building_information, cursor):
        """
        Get the information about the context of a batch of buildings, including the attributes
        of all relations that each building is member of, the attributes of all POIs inside each
        building and all land-use objects that intersect each building.

        Args:
            building_ids (list):
                List of building IDs.
            building_information (dict):
                Dictionary with building IDs as keys and building geometry information as
                values.
            cursor (PostGISDatabase.cursor):
                Connected cursor to the OSM Replication database.

        Returns:
            A dictionary with the building IDs as keys and the building information as values.
        """

        if building_information is None:
            raise ValueError("Geometries of the buildings need to be provided")

        building_vals = ",".join(
            [
                f"({idx}, ST_GeomFromText('{building_information[idx]['geometry_wkt']}', 4326))"
                for idx in building_ids
            ]
        )

        # Get the attributes of intersecting OSM buildings, based on the geometries of the
        # building dataset that is being initialized
        sql_statement = f"""
            WITH buildings (id, geom) AS (
                VALUES {building_vals}
            )
            SELECT buildings.id AS building_id, osm_id, hstore_to_matrix(obp.tags) AS tags
            FROM osm_building_polygons AS obp
            INNER JOIN buildings ON ST_Intersects(buildings.geom, obp.geometry)
            WHERE obp.geometry && buildings.geom
            """
        cursor.execute(sql_statement)
        for building_id, osm_id, tags in cursor:
            building_information[building_id]["relations"].append(
                {"osm_id": osm_id, "tags": dict(tags)}
            )

        # Get the relation attributes of OSM relations that are connected with the OSM
        # buildings intersecting the building dataset being initialized. The relation
        # attributes are retrieved based on the OSM building IDs and the `way` element type of
        # the member (type=1, as defined in the OSM replication database).
        sql_statement = f"""
            WITH buildings (id, geom) AS (
                VALUES {building_vals}
            )
            SELECT
                buildings.id AS building_id,
                obr.osm_id,
                member,
                hstore_to_matrix(obr.tags) AS tags
            FROM osm_building_polygons AS obp
            INNER JOIN buildings ON ST_Intersects(buildings.geom, obp.geometry)
            INNER JOIN osm_building_relations AS obr ON (obr.member = obp.osm_id)
            WHERE obp.geometry && buildings.geom AND obr.type = 1
            """
        cursor.execute(sql_statement)
        for building_id, osm_id, member, tags in cursor:
            building_information[building_id]["relations"].append(
                {"osm_id": osm_id, "tags": dict(tags)}
            )

        # Get the attributes of the OSM POIs inside the buildings
        sql_statement = f"""
            WITH buildings (id, geom) AS (
                VALUES {building_vals}
            )
            SELECT os.key, os.type, buildings.id AS building_id
            FROM osm_spots AS os
            INNER JOIN buildings ON ST_Intersects(buildings.geom, os.geometry)
            WHERE os.geometry && buildings.geom
            """
        cursor.execute(sql_statement)
        for key, value, building_id in cursor:
            building_information[building_id]["osm_spots"].append({key: value})

        # Get the attributes of the OSM land use intersecting the buildings
        sql_statement = f"""
            WITH buildings (id, geom) AS (
                VALUES {building_vals}
            )
            SELECT ol.key, ol.type, buildings.id AS building_id
            FROM osm_lands AS ol
            INNER JOIN buildings ON ST_Intersects(buildings.geom, ol.geometry)
            WHERE buildings.geom && ol.geometry
            """
        cursor.execute(sql_statement)
        for key, value, member in cursor:
            building_information[member]["osm_lands"].append({key: value})
        return building_information

    def get_building_information_query(self, building_ids):
        """
        Create an SQL query that retrieves all attributes related to the buildings themselves.

        Args:
            building_ids (string):
                String-formatted list of building IDs.

        Returns:
            Formatted SQL query.
        """

        height_section = ""
        if self.building_table == "microsoft_buildings":
            height_section = "height,"

        return f"""
            SELECT
                {self.building_id_column},
                {height_section}
                ST_X(ST_Centroid({self.building_geometry_column})) AS longitude,
                ST_Y(ST_Centroid({self.building_geometry_column})) AS latitude,
                ST_AsText({self.building_geometry_column}) AS geometry_wkt,
                ST_Area({self.building_geometry_column}, True) AS area
            FROM {self.building_table}
            WHERE {self.building_id_column} IN ({building_ids})
            """

    @staticmethod
    def get_ghsl_characteristics_type(building_ids, building_information, cursor):
        """
        Intersect the building geometry of a batch of buildings with the GHSL characteristics
        layer and return the type with the largest overlap.

        Args:
            building_ids (list):
                List of building IDs.
            building_information (dict):
                Dictionary with building IDs as keys and building geometry information as
                values.
            cursor (PostGISDatabase.cursor):
                Connected cursor to the OSM Replication database.

        Returns:
            A modified `building_information` dictionary with the GHSL characteristics type
            exhibiting the largest overlap with the buildings.

        """

        # Create a list of buildings with IDs and geometries.
        building_vals = ",".join(
            [
                f"({idx}, ST_GeomFromText('{building_information[idx]['geometry_wkt']}', 4326))"
                for idx in building_ids
            ]
        )

        # Get the total area of the intersection between a GHSL type and the building.
        area_function = """
            SUM(
                ST_Area(
                    ST_Intersection(ghsl_characteristics.geom, ST_Buffer(buildings.geom,0)
                ), True)
            )
            """

        # Select the type with the largest overlap with the buildings. A temporary table
        # `buildings` with the columns `id` and `geom`. The intersection is calculated using the
        # inner statement inside the `FROM` section. Using a combination of the `DISTINCT ON`
        # and the `ORDER BY` statements, it is ensured that the one type with the largest
        # overlap is returned for each building. The `LIMIT` statement in the inner query is
        # added, because the query would not finish otherwise.
        sql_statement = f"""
            WITH buildings (id, geom) AS (
                VALUES {building_vals}
            )
            SELECT DISTINCT ON (id)
                id, type
            FROM (
                SELECT buildings.id, type, {area_function} AS area
                FROM ghsl_characteristics
                INNER JOIN (
                    SELECT id, geom FROM buildings LIMIT {len(building_ids)}
                ) AS buildings ON ghsl_characteristics.geom && buildings.geom
                GROUP BY buildings.id, type
            ) AS ghsl_area
            ORDER BY id, area DESC
            """
        cursor.execute(sql_statement)

        # Add the resulting type to the building information dictionary.
        for building_id, ghsl_characteristics_type in cursor:
            building_information[building_id][
                "ghsl_characteristics_type"
            ] = ghsl_characteristics_type
        return building_information

    @staticmethod
    def batch_iter(iterable, batch_size):
        """
        Returns a batch of features from a list of features.

        Args:
            iterable (iterable):
                List or other iterable containing the features.
            batch_size (int):
                Number of features in one batch.
        """

        batch = []
        for idx, value in enumerate(iterable):
            batch.append(value)
            if (idx + 1) % batch_size == 0:
                yield batch
                batch = []
        if batch:
            yield batch

    def copy_postgis_database_objects(self):
        """
        Get a copy of the `osm_replication_db`, `exposure_db` and the `source_db` objects.

        Returns:
            A copy of each database object, in the order: `osm_replication_db`, `source_db`,
            `exposure_db`.
        """

        osm_replication_db = copy.copy(self.osm_replication_db)
        exposure_db = copy.copy(self.exposure_db)
        source_db = copy.copy(self.source_db)
        return osm_replication_db, source_db, exposure_db
