import yaml
import glob
import argparse
import os
import logging

from databaselib import PostGISDatabase, SpatiaLiteDatabase
from rulelib import Rule

from buildinginitializer import BuildingInitializer, OSMInitializer

logger = logging.getLogger(__name__)


def command_line_interface():
    """
    Command-line interface of the building-initializer. The program first queries (a subset of)
    all building IDs in the source database and processes them in batches using the rules that
    are given as parameters. Optionally, instead of using the command-line options, environment
    variables can be defined as input parameters.

    Options (environment variables in parentheses):
        `-c` (DATABASE_CONFIG_FILE, required):
            Filepath of the config file for accessing the databases.
        `-r` (RULES, required):
            Search pattern to identify all necessary rule filepaths.
        `-b` (BATCH_SIZE, default: 500):
            Optional batch size.
        `-s` (SQL_STATEMENT):
            Optional SQL statement, if a subset of the OSM Replication buildings needs to
            be processed.
        `-p` (NUM_PROCESSORS, default: 1):
            Optional number of processors parameter used for processing the buildings.
        `-t` (BUILDING_TABLE, default: obm_buildings):
            Optional name of the building table.
        `-i` (BUILDING_ID_COLUMN, default: osm_id):
            Optional name of the building ID column.
        `-g` (BUILDING_GEOMETRY_COLUMN, default: geometry):
            Optional name of the geometry column.
    """

    # Set logging configuration
    logging.basicConfig(level=logging.INFO)

    # Create the argument parser and prepare for subparsers
    parser = argparse.ArgumentParser(description="Building-Initializer")
    parser.add_argument(
        "-c",
        "--database-config-file",
        default=os.getenv("DATABASE_CONFIG_FILE", None),
        type=str,
        help="Filepath of the config file for accessing the databases",
    )
    parser.add_argument(
        "-r",
        "--rules",
        default=os.getenv("RULES", None),
        type=str,
        help="Search pattern for rule files",
    )
    parser.add_argument(
        "-b",
        "--batch-size",
        type=int,
        default=os.getenv("BATCH_SIZE", 500),
        help="Optional batch size",
    )
    parser.add_argument(
        "-s",
        "--custom-sql-statement",
        type=str,
        default=os.getenv("CUSTOM_SQL_STATEMENT", None),
        help="Optional SQL statement, if a subset of the OSM Replication buildings needs to be "
        + "processed",
    )
    parser.add_argument(
        "-p",
        "--num-processors",
        type=int,
        default=os.getenv("NUM_PROCESSORS", 1),
        help="Number of processors used for processing the buildings",
    )
    parser.add_argument(
        "-t",
        "--building-table",
        type=str,
        default=os.getenv("BUILDING_TABLE", "obm_buildings"),
        help="Optional name of the building table",
    )
    parser.add_argument(
        "-i",
        "--building-id-column",
        type=str,
        default=os.getenv("BUILDING_ID_COLUMN", "osm_id"),
        help="Optional name of the building ID column",
    )
    parser.add_argument(
        "-g",
        "--building-geometry-column",
        type=str,
        default=os.getenv("BUILDING_GEOMETRY_COLUMN", "geometry"),
        help="Optional name of the geometry column",
    )
    args = parser.parse_args()

    # Because the default parameters are set to the environment variables, we need to check if
    # the required parameters exist
    if not args.database_config_file:
        raise ValueError("Database configuration file is not specified")
    if not args.rules:
        raise ValueError("Rules are not specified")

    # Set attributes from the CLI arguments
    building_table = args.building_table
    building_id_column = args.building_id_column
    building_geometry_column = args.building_geometry_column
    database_config = yaml.safe_load(open(args.database_config_file, "r"))

    # The OSM Replication database has to be a PostGIS database
    osm_replication_db = PostGISDatabase(**database_config["osm_replication"])
    source_db = PostGISDatabase(**database_config["source"])

    # The Building database can be a SpatiaLite or a PostGIS database
    if "filepath" in database_config["exposure"]:
        exposure_db = SpatiaLiteDatabase(database_config["exposure"]["filepath"])
    else:
        exposure_db = PostGISDatabase(**database_config["exposure"])

    # Process rules
    rules = {}
    for rule_filepath in glob.glob(args.rules):
        if rule_filepath.endswith("xml"):
            rule = Rule.load_rule_from_xml(open(rule_filepath, "r").read())
        elif rule_filepath.endswith("zip"):
            rule = Rule.load_rule_from_zip(open(rule_filepath, "rb"))
        else:
            raise ValueError(f"Filepath `{rule_filepath}` does not have a valid file extension")
        rules[rule.name] = rule
    rule_names = "\n\t".join([rule_name for rule_name in rules.keys()])
    logger.info(f"Rules that are used to process buildings:\n\t{rule_names}")

    # If the building table is `obm_buildings`, create an OSMInitializer instance. Otherwise,
    # create a standard BuildingInitializer instance.
    if args.building_table == "obm_buildings":
        # Create OSMInitializer
        building_initializer = OSMInitializer(
            rules=rules,
            osm_replication_db=osm_replication_db,
            exposure_db=exposure_db,
            source_db=source_db,
        )
    else:
        # Create BuildingInitializer
        building_initializer = BuildingInitializer(
            rules=rules,
            osm_replication_db=osm_replication_db,
            exposure_db=exposure_db,
            source_db=source_db,
            building_table=building_table,
            building_id_column=building_id_column,
            building_geometry_column=building_geometry_column,
        )

    # Retrieve building ids.
    building_ids = building_initializer.retrieve_building_ids(args.custom_sql_statement)

    # Run on a single process if nun_processors is 1, otherwise run on multiple processors.
    logger.info(f"Starting the initializer using {args.num_processors} processors")
    if args.num_processors == 1:
        building_initializer.run(building_ids=building_ids, batch_size=args.batch_size)
    else:
        building_initializer.run_multiprocess(
            building_ids=building_ids,
            batch_size=args.batch_size,
            num_processors=args.num_processors,
        )
