import logging

from buildinginitializer import BuildingInitializer

logger = logging.getLogger(__name__)


class OSMInitializer(BuildingInitializer):
    """
    The `BuildingInitializer` class copies and enriches building data from an OpenStreetMap
    replication database.

    Args:
        rules (dict):
            A dictionary with rule names as keys and definitions as values.
        osm_replication_db (PostGISDatabase):
            The `OSM replication` database with the source data.
        exposure_db (AbstractDatabase):
            The output database with processed building information.

    Attributes:
        rule_handler (RuleHandler):
            Evaluates given rules based on a set of arguments.
        osm_replication_db (PostGISDatabase):
            The `OSM replication` database with the source data.
        exposure_db (AbstractDatabase):
            The output database with processed building information.
    """

    building_table: str = "obm_buildings"
    building_id_column: str = "osm_id"
    building_geometry_column: str = "geometry"
    building_source_db_name: str = "osm_replication"

    def __init__(self, rules, osm_replication_db, exposure_db, source_db):
        super().__init__(
            rules=rules,
            osm_replication_db=osm_replication_db,
            exposure_db=exposure_db,
            source_db=source_db,
            building_table=self.building_table,
            building_id_column=self.building_id_column,
            building_geometry_column=self.building_geometry_column,
        )

    def retrieve_building_ids(
        self,
        sql_statement=None,
        building_id_column=None,
        building_table=None,
        building_geometry_column=None,
    ):
        """
        Retrieve a selection of building IDs from the building table, using a custom SQL
        statement. If no statement is given, all IDs are retrieved. Without an SQL statement
        provided, the `building_id_column`, `building_table` and `building_geometry_column`
        are used to create an SQL statement. If these are not given, use the class variables
        instead. The `building_table` is automatically changed to `osm_building_polygons`.

        Args:
            sql_statement (str, optional, default: None):
                The SQL statement used to retrieve the building ids.
            building_id_column (str, optional, default: None):
                ID column name of the building table.
            building_table (str, optional, default: None):
                Name of the building table.
            building_geometry_column (str, optional, default: None):
                Geometry column name of the building table.
        """

        if building_table is None:
            building_table = "osm_building_polygons"

        return super().retrieve_building_ids(
            sql_statement, building_id_column, building_table, building_geometry_column
        )

    def get_additional_information(self, building_ids, building_information, cursor):
        """
        Get the information about the context of a batch of buildings, including the attributes
        of all relations that each building is member of, the attributes of all POIs inside each
        building and all land-use objects that intersect each building.

        Args:
            building_ids (list):
                List of building IDs.
            building_information (dict):
                Dictionary with building IDs as keys and building geometry information as
                values.
            cursor (PostGISDatabase.cursor):
                Connected cursor to the OSM Replication database.

        Returns:
            A dictionary with the building IDs as keys and the building information as values.
        """

        if not isinstance(building_ids, str):
            building_ids = ",".join([str(idx) for idx in building_ids])

        # Get the relation attributes of the buildings by filtering on the building IDs and
        # the `way` element type of the member (type=1, as defined in the OSM replication
        # database).
        sql_statement = f"""
            SELECT {self.building_id_column}, member, hstore_to_matrix(tags) AS tags
            FROM osm_building_relations
            WHERE member IN ({building_ids}) AND type = 1
            """
        cursor.execute(sql_statement)
        for building_id, member, tags in cursor:
            building_information[member]["relations"].append(
                {self.building_id_column: building_id, "tags": dict(tags)}
            )

        # Get the attributes of the POIs inside the buildings
        sql_statement = f"""
            SELECT os.key, os.type, obp.osm_id
            FROM osm_building_polygons AS obp
            INNER JOIN osm_spots AS os ON ST_Intersects(obp.geometry, os.geometry)
            WHERE obp.osm_id IN ({building_ids});
            """
        cursor.execute(sql_statement)
        for k, v, member in cursor:
            building_information[member]["osm_spots"].append({k: v})

        # Get the attributes of the land use intersecting the buildings
        sql_statement = f"""
            SELECT os.key, os.type, obp.osm_id
            FROM osm_building_polygons AS obp
            INNER JOIN osm_lands AS os ON ST_Intersects(os.geometry, obp.geometry)
            WHERE obp.osm_id IN ({building_ids}) AND obp.geometry && os.geometry;
            """
        cursor.execute(sql_statement)
        for k, v, member in cursor:
            building_information[member]["osm_lands"].append({k: v})
        return building_information

    def get_building_information_query(self, building_ids):
        """
        Create an SQL query that retrieves all attributes related to the buildings themselves.

        Args:
            building_ids (string):
                String-formatted list of building IDs.

        Returns:
            Formatted SQL query.
        """

        return f"""
            SELECT
                {self.building_id_column},
                hstore_to_matrix(tags) AS tags,
                ST_X(ST_Centroid({self.building_geometry_column})) AS longitude,
                ST_Y(ST_Centroid({self.building_geometry_column})) AS latitude,
                ST_AsText({self.building_geometry_column}) AS geometry_wkt,
                ST_Area({self.building_geometry_column}, True) AS area
            FROM osm_building_polygons
            WHERE {self.building_id_column} IN ({building_ids})
            """
