#!/usr/bin/env python3

# Copyright (C) 2023:
#   Helmholtz-Zentrum Potsdam Deutsches GeoForschungsZentrum GFZ
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
# General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.

import pytest
import os
import logging
from databaselib import PostGISDatabase, SpatiaLiteDatabase
from rulelib import Rule


logger = logging.getLogger()


def get_env(env_name):
    """
    Returns the environment-variable value given its name.
    """

    env_variable = os.getenv(env_name)
    if env_variable is None:
        raise NameError(
            f"Environment variable `{env_name}` not set. For Gitlab environment "
            f"variables, go to: `Settings` --> `CI/CD` --> `Variables`"
        )
    return env_variable


@pytest.fixture
def source_db():
    """
    Returns a PostGIS database object based on environment variables.
    """

    db_config = {
        "dbname": os.getenv("DB_NAME", None),
        "host": os.getenv("DB_HOST", None),
        "port": os.getenv("DB_PORT", None),
        "username": os.getenv("DB_USERNAME", None),
        "password": os.getenv("DB_PASSWORD", None),
    }

    logger.info(f"Connecting to database {db_config['dbname']}")
    yield PostGISDatabase(**db_config)


@pytest.fixture
def exposure_db():
    """
    Returns a SpatiaLite database object with the `obm_buildings` table.
    """

    yield SpatiaLiteDatabase("./tests/data/exposure.db")


@pytest.fixture
def rules():
    """
    Returns a dictionary with one rule.
    """

    floorspace_rule = Rule.load_rule_from_xml(open("./tests/data/storeys.xml", "r").read())
    yield {floorspace_rule.name: floorspace_rule}
