#!/usr/bin/env python3

# Copyright (C) 2023:
#   Helmholtz-Zentrum Potsdam Deutsches GeoForschungsZentrum GFZ
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
# General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.

import logging
import pytest

from psycopg2 import ProgrammingError

from buildinginitializer import OSMInitializer

logger = logging.getLogger()


def test_get_building_information(source_db, exposure_db):
    """
    Test the retrieval of building information of the `OSMInitializer` class on the
    micro-OSM Replication database with data located at the Telegrafenberg campus in Potsdam,
    Germany.
    """

    building_initializer = OSMInitializer(
        {}, source_db=source_db, osm_replication_db=source_db, exposure_db=exposure_db
    )
    source_db.connect()
    building_ids = [352224937, 173039036, 27139591]
    buildings = building_initializer.get_building_information(building_ids, source_db.cursor)
    buildings = building_initializer.get_additional_information(
        building_ids, buildings, source_db.cursor
    )
    buildings = building_initializer.get_ghsl_characteristics_type(
        building_ids, buildings, source_db.cursor
    )
    source_db.close()

    # Einsteinturm
    assert buildings[27139591] is not None
    assert buildings[27139591]["tags"]["building:levels"] == "4"
    assert buildings[27139591]["tags"]["building:material"] == "brick;plaster"
    assert len(buildings[27139591]["relations"]) != 0
    assert -8907330 in [relation["osm_id"] for relation in buildings[27139591]["relations"]]
    assert buildings[27139591]["ghsl_characteristics_type"] == 11

    # Refractor GFZ
    assert buildings[352224937] is not None
    assert buildings[352224937]["tags"]["height"] == "17"
    assert buildings[352224937]["tags"]["architect"] == "Paul Emanuel Spieker"
    assert buildings[27139591]["ghsl_characteristics_type"] == 11

    # Building A70
    assert buildings[173039036] is not None
    assert buildings[173039036]["tags"]["building"] == "office"
    assert buildings[173039036]["tags"]["building:levels"] == "4"
    assert buildings[27139591]["ghsl_characteristics_type"] == 11


def test_evaluate_building_information(rules, source_db, exposure_db):
    """
    Test the evaluation of building information of the `OSMInitializer` class on the
    micro-OSM Replication database with data located at the Telegrafenberg campus in Potsdam,
    Germany.
    """
    logging.basicConfig(level=logging.INFO)

    building_initializer = OSMInitializer(
        rules, source_db=source_db, osm_replication_db=source_db, exposure_db=exposure_db
    )
    source_db.connect()
    building_ids = [352224937, 173039036, 27139591]
    buildings = building_initializer.get_building_information(building_ids, source_db.cursor)
    buildings = building_initializer.get_additional_information(
        building_ids, buildings, source_db.cursor
    )
    source_db.close()
    output = building_initializer.evaluate_building_information(buildings)
    assert output[173039036]["storeys"] == 4
    assert output[352224937]["storeys"] == 2
    assert output[27139591]["storeys"] == 4


def test_upsert_building_information(source_db, exposure_db):
    """
    Test if the `upsert` is working. Regardless whether the given OSM IDs are already in the
    building database, the new values should be updated.
    """

    building_initializer = OSMInitializer(
        {}, source_db=source_db, osm_replication_db=source_db, exposure_db=exposure_db
    )
    building_initializer.outputs = {"occupancy", "storeys", "relation_id"}
    building_information = {
        173039036: {"occupancy": "TEST-1", "storeys": 10, "relation_id": None},
        352224937: {"storeys": 8, "relation_id": 5, "occupancy": "TEST-2"},
    }
    exposure_db.connect()
    building_initializer.upsert(building_information, exposure_db.cursor)
    exposure_db.cursor.execute(
        """
        SELECT osm_id, occupancy, storeys, relation_id
        FROM obm_buildings
        WHERE osm_id IN (173039036,352224937)
        """
    )
    for building_id, occupancy, storeys, relation_id in exposure_db.cursor:
        if building_id == 173039036:
            assert occupancy == "TEST-1"
            assert storeys == 10
            assert relation_id is None
        elif building_id == 352224937:
            assert occupancy == "TEST-2"
            assert storeys == 8
            assert relation_id == 5
        else:
            assert False
    exposure_db.close()


def test_run(source_db, exposure_db, rules):
    """
    Test the `OSMInitializer.run()` function.
    """

    building_initializer = OSMInitializer(
        rules, source_db=source_db, osm_replication_db=source_db, exposure_db=exposure_db
    )
    building_ids = [352224937, 173039036, 27139591]
    building_initializer.run(building_ids=building_ids, batch_size=4)
    exposure_db.connect()
    exposure_db.cursor.execute(
        """
        SELECT osm_id, storeys
        FROM obm_buildings
        WHERE osm_id IN (173039036,352224937)
        """
    )
    for building_id, storeys in exposure_db.cursor:
        if building_id == 173039036:
            assert storeys == 4
        elif building_id == 352224937:
            assert storeys == 2
        else:
            assert False
    exposure_db.close()


def test_retrieve_building_ids(source_db, exposure_db, rules):
    """
    Test the `OSMInitializer.retrieve_building_ids()` function. The first test checks if
    the standard function returns all building IDs currently existing in the database. The
    second test checks the use of a custom SQL statement. The last test checks if providing an
    incorrect table name raises a Psycopg `ProgrammingError` exception.
    """

    building_initializer = OSMInitializer(
        {}, source_db=source_db, osm_replication_db=source_db, exposure_db=exposure_db
    )

    building_ids = building_initializer.retrieve_building_ids()
    assert 352224937 in building_ids
    assert 173039036 in building_ids
    assert 27139591 in building_ids
    assert len(building_ids) == 91  # All buildings on the Telegrafenberg.

    sql_statement = """
        SELECT osm_id
        FROM osm_building_polygons
        WHERE osm_id IN (352224937, 173039036)
        """

    building_ids = building_initializer.retrieve_building_ids(sql_statement=sql_statement)
    assert 352224937 in building_ids
    assert 173039036 in building_ids
    assert 27139591 not in building_ids
    assert len(building_ids) == 2

    with pytest.raises(ProgrammingError) as exception_information:
        building_initializer.retrieve_building_ids(building_table="non_existent_table")
        assert "relation * does not exist " in str(exception_information.value)


@pytest.mark.xfail(raises=TypeError)
def test_multiprocessing(source_db, exposure_db, rules):
    """
    Multiprocessing cannot be tested, because the building database is a SpatiaLite database.
    Instead, test if running the `run_multiprocess` function leads to an error.
    """

    building_initializer = OSMInitializer(
        rules, source_db=source_db, osm_replication_db=source_db, exposure_db=exposure_db
    )
    building_ids = [352224937, 173039036, 27139591]
    building_initializer.run_multiprocess(
        building_ids=building_ids, batch_size=1, num_processors=1
    )
